
from .interface import run, run_async # noqa
from .exceptions import AnsibleRunnerException, ConfigurationError, CallbackError # noqa
from .runner_config import RunnerConfig # noqa
from .runner import Runner # noqa



from uuid import uuid4
import pkg_resources
import threading
import argparse
import signal
import errno
import json
import stat
import sys
import os

from .runner_config import RunnerConfig
from .runner import Runner
from .utils import to_artifacts


def run(**kwargs):
    '''
    Run an Ansible Runner task in the foreground and return a Runner object when complete.

    Args:

        private_data_dir (string, path): The directory containing all runner metadata needed
            to invoke the runner module

        ident (string, optional): The run identifier for this invocation of Runner. Will be used
            to create and name the artifact directory holding the results of the invocation

        playbook (string, filename or list): The playbook relative path located in the private_data_dir/project
            directory that will be invoked by runner when executing Ansible.  If this value is provided as a
            Python list object, the playbook will be written to disk and then executed.

        inventory (string): Override the inventory directory/file supplied with runner metadata at
            private_data_dir/inventory with a specific list of hosts.  This kwarg accepts either a
            full path to the inventory file in the private_data_dir, a native Python dict supporting
            YAML inventory structure or a text INI formatted string.

        envvars (dict, optional): Any environment variables to be used when running Ansible.

        extravars (dict, optional): Any extra variables to be passed to Ansible at runtime using
            the -e option when calling ansible-playbook

        passwords (dict, optional): A dict object that contains password prompt patterns and response
            values used when processing output from ansible-playbook

        settings (dict, optional): A dict objec that contains values for ansible-runner runtime
            settings.

        ssh_key (string, optional): The ssh private key passed to ssh-agent as part of the
            ansible-playbook run

        limit (string, optional): Matches ansible's --limit parameter to further constrain the inventory to be used

    Returns:
        Runner: An object that holds details and results from the invocation of Ansible itself
    '''
    to_artifacts(kwargs)
    rc = RunnerConfig(**kwargs)
    rc.prepare()
    r = Runner(rc)
    r.run()
    return r


def run_async(**kwargs):
    '''
    Run an Ansible Runner task in the background and return a thread object and  Runner object when complete.

    Args:

        private_data_dir (string, path): The directory containing all runner metadata needed
            to invoke the runner module

        ident (string, optional): The run identifier for this invocation of Runner. Will be used
            to create and name the artifact directory holding the results of the invocation

        playbook (string, filename or list): The playbook relative path located in the private_data_dir/project
            directory that will be invoked by runner when executing Ansible.  If this value is provided as a
            Python list object, the playbook will be written to disk and then executed.

        inventory (string): Override the inventory directory/file supplied with runner metadata at
            private_data_dir/inventory with a specific list of hosts.  This kwarg accepts either a
            full path to the inventory file in the private_data_dir, a native Python dict supporting
            YAML inventory structure or a text INI formatted string.

        envvars (dict, optional): Any environment variables to be used when running Ansible.

        extravars (dict, optional): Any extra variables to be passed to Ansible at runtime using
            the -e option when calling ansible-playbook

        passwords (dict, optional): A dict object that contains password prompt patterns and response
            values used when processing output from ansible-playbook

        settings (dict, optional): A dict objec that contains values for ansible-runner runtime
            settings.

        ssh_key (string, optional): The ssh private key passed to ssh-agent as part of the
            ansible-playbook run

        limit (string, optional): Matches ansible's --limit parameter to further constrain the inventory to be used

    Returns:
        threadObj, Runner: An object representing the thread itself and a Runner instance that holds details
        and results from the invocation of Ansible itself
    '''
    to_artifacts(kwargs)
    rc = RunnerConfig(**kwargs)
    rc.prepare()
    r = Runner(rc)
    runner_thread = threading.Thread(target=r.run)
    runner_thread.start()
    return runner_thread, r



def main():
    version = pkg_resources.require("ansible_runner")[0].version
    parser = argparse.ArgumentParser(description='manage ansible execution')
    parser.add_argument('--version', action='version', version=version)
    parser.add_argument('command', choices=['run', 'start',
                                            'stop', 'is-alive'])
    parser.add_argument('private_data_dir',
                        help='Base directory containing Runner metadata (project, inventory, etc')
    parser.add_argument("--hosts")
    parser.add_argument("-p", "--playbook", default=os.getenv("RUNNER_PLAYBOOK", None))
    parser.add_argument("-i", "--ident",
                        default=uuid4(),
                        help="An identifier that will be used when generating the"
                             "artifacts directory and can be used to uniquely identify a playbook run")
    args = parser.parse_args()
    pidfile = os.path.join(args.private_data_dir, 'pid')
    try:
        os.makedirs(args.private_data_dir)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(args.private_data_dir):
            pass
        else:
            raise
    stderr_path = os.path.join(args.private_data_dir, 'daemon.log')
    if not os.path.exists(stderr_path):
        os.mknod(stderr_path, stat.S_IFREG | stat.S_IRUSR | stat.S_IWUSR)
    stderr = open(stderr_path, 'w+')

    if args.command in ('start', 'run'):
        if args.command == 'start':
            import daemon
            from daemon.pidfile import TimeoutPIDLockFile
            context = daemon.DaemonContext(
                pidfile=TimeoutPIDLockFile(pidfile),
                stderr=stderr
            )
        else:
            context = threading.Lock()
        with context:
            run_options = dict(private_data_dir=args.private_data_dir,
                               ident=args.ident,
                               playbook=args.playbook)
            if args.hosts is not None:
                run_options.update(inventory=args.hosts)
            run(**run_options)
            sys.exit(0)
    try:
        with open(pidfile, 'r') as f:
            pid = int(f.readline())
    except IOError:
        sys.exit(1)

    if args.command == 'stop':
        try:
            with open(os.path.join(args.private_data_dir, 'args'), 'r') as args:
                Runner.handle_termination(pid, json.load(args), 'bwrap')
        except IOError:
            Runner.handle_termination(pid, [], 'bwrap')
    elif args.command == 'is-alive':
        try:
            os.kill(pid, signal.SIG_DFL)
            sys.exit(0)
        except OSError:
            sys.exit(1)



from StringIO import StringIO
import base64
import json
import sys
import re
import os
import fcntl
import tempfile
import hashlib

from collections import Iterable, Mapping

from six import string_types



def isplaybook(obj):
    '''
    Inspects the object and returns if it is a playbook

    Args:
        obj (object): The object to be inspected by this function

    Returns:
        boolean: True if the object is a list and False if it is not
    '''
    return isinstance(obj, Iterable) and (not isinstance(obj, string_types) and not isinstance(obj, Mapping))


def isinventory(obj):
    '''
    Inspects the object and returns if it is an inventory

    Args:
        obj (object): The object to be inspected by this function

    Returns:
        boolean: True if the object is an inventory dict and False if it is not
    '''
    return isinstance(obj, Mapping) or isinstance(obj, string_types)


def dump_artifact(obj, path, filename=None):
    '''
    Write the artifact to disk at the specified path

    Args:
        obj (string): The string object to be dumped to disk in the specified
            path.  The artifact filename will be automatically created

        path (string): The full path to the artifacts data directory.

        filename (string, optional): The name of file to write the artifact to.
            If the filename is not provided, then one will be generated.

    Returns:
        string: The full path filename for the artifact that was generated
    '''
    p_sha1 = None

    if not os.path.exists(path):
        os.makedirs(path)
    else:
        p_sha1 = hashlib.sha1()
        p_sha1.update(obj)

    if filename is None:
        fd, fn = tempfile.mkstemp(dir=path)
    else:
        fn = os.path.join(path, filename)

    if os.path.exists(fn):
        c_sha1 = hashlib.sha1()
        c_sha1.update(open(fn).read())

    if not os.path.exists(fn) or p_sha1.hexdigest() != c_sha1.hexdigest():
        lock_fp = os.path.join(path, '.artifact_write_lock')
        lock_fd = os.open(lock_fp, os.O_RDWR | os.O_CREAT, 0o600)
        fcntl.lockf(lock_fd, fcntl.LOCK_EX)

        try:
            with open(fn, 'w') as f:
                f.write(str(obj))
        finally:
            fcntl.lockf(lock_fd, fcntl.LOCK_UN)

    return fn


def to_artifacts(kwargs):
    '''
    Introspect the kwargs and dump objects to disk
    '''
    private_data_dir = kwargs.get('private_data_dir')
    if not private_data_dir:
        private_data_dir = tempfile.mkdtemp()
        kwargs['private_data_dir'] = private_data_dir

    if not os.path.exists(private_data_dir):
        raise ValueError('private_data_dir path is either invalid or does not exist')

    for key in ('playbook', 'inventory'):
        obj = kwargs.get(key)
        if obj:
            if key == 'playbook' and isplaybook(obj):
                path = os.path.join(private_data_dir, 'project')
                kwargs['playbook'] = dump_artifact(json.dumps(obj), path, 'main.json')

            elif key == 'inventory' and isinventory(obj):
                path = os.path.join(private_data_dir, 'inventory')
                if isinstance(obj, Mapping):
                    kwargs['inventory'] = dump_artifact(json.dumps(obj), path, 'hosts.json')
                elif isinstance(obj, string_types):
                    if not os.path.exists(obj):
                        kwargs['inventory'] = dump_artifact(obj, path, 'hosts')

    for key in ('envvars', 'extravars', 'passwords', 'settings'):
        obj = kwargs.get(key)
        if obj:
            path = os.path.join(private_data_dir, 'env')
            dump_artifact(json.dumps(obj), path, key)
            kwargs.pop(key)

    if 'ssh_key' in kwargs:
        path = os.path.join(private_data_dir, 'env')
        dump_artifact(str(kwargs['ssh_key']), path, 'ssh_key')
        kwargs.pop('ssh_key')


class OutputEventFilter(object):
    '''
    File-like object that looks for encoded job events in stdout data.
    '''

    EVENT_DATA_RE = re.compile(r'\x1b\[K((?:[A-Za-z0-9+/=]+\x1b\[\d+D)+)\x1b\[K')

    def __init__(self, handle, event_callback):
        self._event_callback = event_callback
        self._event_ct = 0
        self._counter = 1
        self._start_line = 0
        self._handle = handle
        self._buffer = StringIO()
        self._last_chunk = ''
        self._current_event_data = None

    def flush(self):
        # pexpect wants to flush the file it writes to, but we're not
        # actually capturing stdout to a raw file; we're just
        # implementing a custom `write` method to discover and emit events from
        # the stdout stream
        pass

    def write(self, data):
        self._buffer.write(data)

        # keep a sliding window of the last chunk written so we can detect
        # event tokens and determine if we need to perform a search of the full
        # buffer
        should_search = '\x1b[K' in (self._last_chunk + data)
        self._last_chunk = data

        # Only bother searching the buffer if we recently saw a start/end
        # token (\x1b[K)
        while should_search:
            value = self._buffer.getvalue()
            match = self.EVENT_DATA_RE.search(value)
            if not match:
                break
            try:
                base64_data = re.sub(r'\x1b\[\d+D', '', match.group(1))
                event_data = json.loads(base64.b64decode(base64_data))
            except ValueError:
                event_data = {}
            event_data = self._emit_event(value[:match.start()], event_data)
            stdout_actual = event_data['stdout'] if 'stdout' in event_data else None
            remainder = value[match.end():]
            self._buffer = StringIO()
            self._buffer.write(remainder)

            if stdout_actual:
                sys.stdout.write(stdout_actual + "\n")
                self._handle.write(stdout_actual + "\n")

            self._last_chunk = remainder

    def close(self):
        value = self._buffer.getvalue()
        if value:
            self._emit_event(value)
            self._buffer = StringIO()
        self._event_callback(dict(event='EOF'))

    def _emit_event(self, buffered_stdout, next_event_data=None):
        next_event_data = next_event_data or {}
        if self._current_event_data:
            event_data = self._current_event_data
            stdout_chunks = [buffered_stdout]
        elif buffered_stdout:
            event_data = dict(event='verbose')
            stdout_chunks = buffered_stdout.splitlines(True)
        else:
            event_data = dict()
            stdout_chunks = []

        for stdout_chunk in stdout_chunks:
            event_data['counter'] = self._counter
            self._counter += 1
            event_data['stdout'] = stdout_chunk[:-2] if len(stdout_chunk) > 2 else ""
            n_lines = stdout_chunk.count('\n')
            event_data['start_line'] = self._start_line
            event_data['end_line'] = self._start_line + n_lines
            self._start_line += n_lines
            if self._event_callback:
                self._event_callback(event_data)
                self._event_ct += 1
        if next_event_data.get('uuid', None):
            self._current_event_data = next_event_data
        else:
            self._current_event_data = None
        return event_data



import os
import stat
import time
import json
import errno
import signal
import codecs
import collections

import pexpect
import psutil

from .utils import OutputEventFilter
from .exceptions import CallbackError, AnsibleRunnerException


class Runner(object):

    def __init__(self, config, cancel_callback=None, remove_partials=True):
        self.config = config
        self.cancel_callback = cancel_callback
        self.canceled = False
        self.timed_out = False
        self.errored = False
        self.status = "unstarted"
        self.rc = None
        self.remove_partials = remove_partials

    def event_callback(self, event_data):
        '''
        Invoked for every Ansible event to collect stdout with the event data and store it for
        later use
        '''
        if 'uuid' in event_data:
            filename = '{}-partial.json'.format(event_data['uuid'])
            partial_filename = os.path.join(self.config.artifact_dir,
                                            'job_events',
                                            filename)
            full_filename = os.path.join(self.config.artifact_dir,
                                         'job_events',
                                         '{}-{}.json'.format(event_data['counter'],
                                                             event_data['uuid']))
            try:
                with codecs.open(partial_filename, 'r', encoding='utf-8') as read_file:
                    partial_event_data = json.load(read_file)
                event_data.update(partial_event_data)
                with codecs.open(full_filename, 'w', encoding='utf-8') as write_file:
                    json.dump(event_data, write_file)
                if self.remove_partials:
                    os.remove(partial_filename)
            except IOError as e:
                print("Failed writing event data: {}".format(e))

    def run(self):
        '''
        Launch the Ansible task configured in self.config (A RunnerConfig object), returns once the
        invocation is complete
        '''
        self.status = "starting"
        stdout_filename = os.path.join(self.config.artifact_dir, 'stdout')

        try:
            os.makedirs(self.config.artifact_dir)
            os.mknod(stdout_filename, stat.S_IFREG | stat.S_IRUSR | stat.S_IWUSR)
        except OSError as exc:
            if exc.errno == errno.EEXIST and os.path.isdir(self.config.artifact_dir):
                pass
            else:
                raise

        stdout_handle = codecs.open(stdout_filename, 'w', encoding='utf-8')
        stdout_handle = OutputEventFilter(stdout_handle, self.event_callback)

        if not isinstance(self.config.expect_passwords, collections.OrderedDict):
            # We iterate over `expect_passwords.keys()` and
            # `expect_passwords.values()` separately to map matched inputs to
            # patterns and choose the proper string to send to the subprocess;
            # enforce usage of an OrderedDict so that the ordering of elements in
            # `keys()` matches `values()`.
            expect_passwords = collections.OrderedDict(self.config.expect_passwords)
        password_patterns = expect_passwords.keys()
        password_values = expect_passwords.values()

        self.status = 'running'
        child = pexpect.spawn(
            self.config.command[0],
            self.config.command[1:],
            cwd=self.config.cwd,
            env=self.config.env,
            ignore_sighup=True,
            encoding='utf-8',
            echo=False,
        )
        child.logfile_read = stdout_handle
        last_stdout_update = time.time()

        job_start = time.time()
        while child.isalive():
            result_id = child.expect(password_patterns,
                                     timeout=self.config.pexpect_timeout,
                                     searchwindowsize=100)
            password = password_values[result_id]
            if password is not None:
                child.sendline(password)
                last_stdout_update = time.time()
            if self.cancel_callback:
                try:
                    self.canceled = self.cancel_callback()
                except Exception:
                    # TODO: logger.exception('Could not check cancel callback - cancelling immediately')
                    #if isinstance(extra_update_fields, dict):
                    #    extra_update_fields['job_explanation'] = "System error during job execution, check system logs"
                    raise CallbackError("Exception in Cancel Callback")
            if not self.canceled and self.config.job_timeout != 0 and (time.time() - job_start) > self.config.job_timeout:
                self.timed_out = True
                # if isinstance(extra_update_fields, dict):
                #     extra_update_fields['job_explanation'] = "Job terminated due to timeout"
            if self.canceled or self.timed_out or self.errored:
                # TODO: proot_cmd
                Runner.handle_termination(child.pid, child.args, proot_cmd=None, is_cancel=self.canceled)
            if self.config.idle_timeout and (time.time() - last_stdout_update) > self.config.idle_timeout:
                child.close(True)
                self.canceled = True

        if self.canceled:
            self.status = "canceled"
        elif child.exitstatus == 0 and not self.timed_out:
            self.status = "successful"
        else:
            self.status = "failed"
        self.rc = child.exitstatus

        for filename, data in [
            ('status', self.status),
            ('rc', self.rc),
        ]:
            artifact_path = os.path.join(self.config.artifact_dir, filename)
            if not os.path.exists(artifact_path):
                os.mknod(artifact_path, stat.S_IFREG | stat.S_IRUSR | stat.S_IWUSR)
            with open(artifact_path, 'w') as f:
                f.write(str(data))
        return self.status, self.rc

    @property
    def stdout(self):
        '''
        Returns an open file handle to the stdout representing the Ansible run
        '''
        stdout_path = os.path.join(self.config.artifact_dir, 'stdout')
        if not os.path.exists(stdout_path):
            raise AnsibleRunnerException("stdout missing")
        return open(os.path.join(self.config.artifact_dir, 'stdout'), 'r')

    @property
    def events(self):
        '''
        A generator that will return all ansible job events in the order that they were emitted from Ansible

        Example:

            {  
               "event":"runner_on_ok",
               "uuid":"00a50d9c-161a-4b74-b978-9f60becaf209",
               "stdout":"ok: [localhost] => {\\r\\n    \\"   msg\\":\\"Test!\\"\\r\\n}",
               "counter":6,
               "pid":740,
               "created":"2018-04-05T18:24:36.096725",
               "end_line":10,
               "start_line":7,
               "event_data":{  
                  "play_pattern":"all",
                  "play":"all",
                  "task":"debug",
                  "task_args":"msg=Test!",
                  "remote_addr":"localhost",
                  "res":{  
                     "msg":"Test!",
                     "changed":false,
                     "_ansible_verbose_always":true,
                     "_ansible_no_log":false
                  },
                  "pid":740,
                  "play_uuid":"0242ac11-0002-443b-cdb1-000000000006",
                  "task_uuid":"0242ac11-0002-443b-cdb1-000000000008",
                  "event_loop":null,
                  "playbook_uuid":"634edeee-3228-4c17-a1b4-f010fdd42eb2",
                  "playbook":"test.yml",
                  "task_action":"debug",
                  "host":"localhost",
                  "task_path":"/tmp/demo/project/test.yml:3"
               }
           }
        '''
        event_path = os.path.join(self.config.artifact_dir, 'job_events')
        if not os.path.exists(event_path):
            raise AnsibleRunnerException("events missing")
        dir_events = os.listdir(event_path)
        dir_events.sort(lambda x, y: int(x.split("-", 1)[0]) - int(y.split("-", 1)[0]))
        for event_file in dir_events:
            with codecs.open(os.path.join(event_path, event_file), 'r', encoding='utf-8') as event_file_actual:
                event = json.load(event_file_actual)
            yield event

    @property
    def stats(self):
        '''
        Returns the final high level stats from the Ansible run

        Example:
            {'dark': {}, 'failures': {}, 'skipped': {}, 'ok': {u'localhost': 2}, 'processed': {u'localhost': 1}}
        '''
        last_event = filter(lambda x: 'event' in x and x['event'] == 'playbook_on_stats',
                            self.events)
        if not last_event:
            return None
        last_event = last_event[0]['event_data']
        return dict(skipped=last_event['skipped'],
                    ok=last_event['ok'],
                    dark=last_event['dark'],
                    failures=last_event['failures'],
                    processed=last_event['processed'])

    def host_events(self, host):
        '''
        Given a host name, this will return all task events executed on that host
        '''
        all_host_events = filter(lambda x: 'event_data' in x and 'host' in x['event_data'] and x['event_data']['host'] == host,
                                 self.events)
        return all_host_events

    @classmethod
    def handle_termination(pid, args, proot_cmd, is_cancel=True):
        '''
        Internal method to terminate a subprocess spawned by `pexpect` representing an invocation of runner.

        :param pid:       the process id of the running the job.
        :param args:      the args for the job, i.e., ['ansible-playbook', 'abc.yml']
        :param proot_cmd  the command used to isolate processes i.e., `bwrap`
        :param is_cancel: flag showing whether this termination is caused by
                          instance's cancel_flag.
        '''
        try:
            if proot_cmd in ' '.join(args):
                if not psutil:
                    os.kill(pid, signal.SIGKILL)
                else:
                    try:
                        main_proc = psutil.Process(pid=pid)
                        child_procs = main_proc.children(recursive=True)
                        for child_proc in child_procs:
                            os.kill(child_proc.pid, signal.SIGKILL)
                        os.kill(main_proc.pid, signal.SIGKILL)
                    except (TypeError, psutil.Error):
                        os.kill(pid, signal.SIGKILL)
            else:
                os.kill(pid, signal.SIGTERM)
            time.sleep(3)
        except OSError:
            raise
            #keyword = 'cancel' if is_cancel else 'timeout'
            #TODO: logger.warn("Attempted to %s already finished job, ignoring" % keyword)



#!/usr/bin/env python

# Copyright (c) 2018 Red Hat, Inc.
# All Rights Reserved.

from setuptools import setup, find_packages

setup(
    name="ansible-runner",
    version="1.0.1",
    author='Red Hat Ansible',
    url="https://github.com/ansible/ansible-runner",
    packages=find_packages(),
    install_requires=[
        'psutil',
        'pexpect',
        'python-daemon',
        'PyYAML',
    ],
    zip_safe=False,
    entry_points={
        'console_scripts': [
            'ansible-runner=ansible_runner.interface:main',
        ],
    },
)



import os
import re
import yaml
import pipes
import thread
import pexpect
from uuid import uuid4

from .exceptions import ConfigurationError


class RunnerConfig(object):

    def __init__(self,
                 private_data_dir=None, playbook=None, ident=uuid4(),
                 inventory=None, limit=None,
                 module=None, module_args=None):
        self.private_data_dir = os.path.abspath(private_data_dir)
        self.ident = ident
        self.playbook = playbook
        self.inventory = inventory
        self.limit = limit
        self.module = module
        self.module_args = module_args
        if self.ident is None:
            self.artifact_dir = os.path.join(self.private_data_dir, "artifacts")
        else:
            self.artifact_dir = os.path.join(self.private_data_dir, "artifacts", "{}".format(self.ident))

    def prepare_inventory(self):
        if self.inventory is None:
            self.inventory  = os.path.join(self.private_data_dir, "inventory")

    def prepare_env(self):
        try:
            with open(os.path.join(self.private_data_dir, "env", "passwords"), 'r') as f:
                self.expect_passwords = {
                    re.compile(pattern, re.M): password
                    for pattern, password in yaml.safe_load(f.read()).items()
                }
        except Exception:
            # TODO: logging
            print("Not loading passwords")
            self.expect_passwords = dict()
        self.expect_passwords[pexpect.TIMEOUT] = None
        self.expect_passwords[pexpect.EOF] = None

        try:
            with open(os.path.join(self.private_data_dir, "env", "envvars"), 'r') as f:
                self.env = yaml.safe_load(f.read())
        except Exception:
            print("Not loading environment vars")
            self.env = dict()

        try:
            with open(os.path.join(self.private_data_dir, "env", "extravars"), 'r') as f:
                self.extra_vars = yaml.safe_load(f.read())
        except Exception:
            print("Not loading extra vars")
            self.extra_vars = dict()

        try:
            with open(os.path.join(self.private_data_dir, "env", "settings"), 'r') as f:
                self.settings = yaml.safe_load(f.read())
        except Exception:
            print("Not loading settings")
            self.settings = dict()

        try:
            with open(os.path.join(self.private_data_dir, "env", "ssh_key"), 'r') as f:
                self.ssh_key_data = f.read()
        except Exception:
            print("Not loading ssh key")
            self.ssh_key_data = None

        self.idle_timeout = self.settings.get('idle_timeout', 120)
        self.job_timeout = self.settings.get('job_timeout', 120)
        self.pexpect_timeout = self.settings.get('pexpect_timeout', 5)

        if 'AD_HOC_COMMAND_ID' in self.env:
            self.cwd = self.private_data_dir
        else:
            self.cwd = os.path.join(self.private_data_dir, 'project')

    def prepare_command(self):
        if os.path.exists(os.path.join(self.private_data_dir, 'args')):
            with open(os.path.join(self.private_data_dir, 'args'), 'r') as args:
                self.command = yaml.safe_load(args)
        else:
            self.command = self.generate_ansible_command()

    def prepare(self):
        if self.private_data_dir is None:
            raise ConfigurationError("Runner Base Directory is not defined")
        if self.playbook is None: # TODO: ad-hoc mode, module and args
            raise ConfigurationError("Runner playbook is not defined")
        if not os.path.exists(self.artifact_dir):
            os.makedirs(self.artifact_dir)

        self.prepare_inventory()
        self.prepare_env()
        self.prepare_command()

        # write the SSH key data into a fifo read by ssh-agent
        if self.ssh_key_data:
            self.ssh_key_path = os.path.join(self.artifact_dir, 'ssh_key_data')
            self.ssh_auth_sock = os.path.join(self.artifact_dir, 'ssh_auth.sock')
            self.open_fifo_write(self.ssh_key_path, self.ssh_key_data)
            self.command = self.wrap_args_with_ssh_agent(self.command, self.ssh_key_path, self.ssh_auth_sock)

        # Use local callback directory
        callback_dir = os.getenv('AWX_LIB_DIRECTORY')
        if callback_dir is None:
            callback_dir = os.path.join(os.path.split(os.path.abspath(__file__))[0],
                                        "callbacks")
        self.env['ANSIBLE_CALLBACK_PLUGINS'] = callback_dir
        if 'AD_HOC_COMMAND_ID' in self.env:
            self.env['ANSIBLE_STDOUT_CALLBACK'] = 'minimal'
        else:
            self.env['ANSIBLE_STDOUT_CALLBACK'] = 'awx_display'
        self.env['ANSIBLE_RETRY_FILES_ENABLED'] = 'False'
        self.env['ANSIBLE_HOST_KEY_CHECKING'] = 'False'
        self.env['AWX_ISOLATED_DATA_DIR'] = self.artifact_dir
        self.env['PYTHONPATH'] = self.env.get('PYTHONPATH', '') + callback_dir + ':'


    def generate_ansible_command(self):
        if self.module is not None:
            base_command = 'ansible'
        else:
            base_command = 'ansible-playbook'
        exec_list = [base_command]
        exec_list.append("-i")
        exec_list.append(self.inventory)
        if self.limit is not None:
            exec_list.append("--limit")
            exec_list.append(self.limit)
        if self.extra_vars:
            for evar in self.extra_vars:
                exec_list.append("-e")
                exec_list.append("{}={}".format(evar, self.extra_vars[evar]))
        # Other parameters
        if base_command.endswith('ansible-playbook'):
            exec_list.append(self.playbook)
        elif base_command == 'ansible':
            exec_list.append("-m")
            exec_list.append(self.module)
            if self.module_args is not None:
                exec_list.append("-a")
                exec_list.append(self.module_args)
        return exec_list


    def wrap_args_with_ssh_agent(self, args, ssh_key_path, ssh_auth_sock=None, silence_ssh_add=False):
        if ssh_key_path:
            ssh_add_command = self.args2cmdline('ssh-add', ssh_key_path)
            if silence_ssh_add:
                ssh_add_command = ' '.join([ssh_add_command, '2>/dev/null'])
            cmd = ' && '.join([ssh_add_command,
                               self.args2cmdline('rm', '-f', ssh_key_path),
                               self.args2cmdline(*args)])
            args = ['ssh-agent']
            if ssh_auth_sock:
                args.extend(['-a', ssh_auth_sock])
            args.extend(['sh', '-c', cmd])
        return args


    def open_fifo_write(self, path, data):
        # TODO: Switch to utility function
        '''open_fifo_write opens the fifo named pipe in a new thread.
        This blocks the thread until an external process (such as ssh-agent)
        reads data from the pipe.
        '''
        os.mkfifo(path, 0o600)
        thread.start_new_thread(lambda p, d: open(p, 'w').write(d), (path, data))


    def args2cmdline(self, *args):
        # TODO: switch to utility function
        return ' '.join([pipes.quote(a) for a in args])
